import os
from contextlib import contextmanager
from typing import Type, Tuple

from yaml import SafeLoader, BaseLoader
from entrypoints import get_group_all as get_entrypoints

from ..config.yaml import (
    load_configuration as load_yaml_configuration,
    yaml_constructor,
)
from ..config.python import load_configuration as load_python_configuration
from ..config.mapping import Translator, SectionPlugin


class COBalDLoader(SafeLoader):
    """Loader with access to COBalD configuration constructors"""


def add_constructor_plugins(entry_point_group: str, loader: Type[BaseLoader]) -> None:
    """
    Add PyYAML constructors from an entry point group to a loader

    :param loader: the PyYAML loader which uses the plugins
    :param entry_point_group: entry point group to search

    .. note::

        This directly modifies the ``loader`` by
        calling :py:meth:`~.BaseLoader.add_constructor`.
    """
    for entry in get_entrypoints(entry_point_group):
        if entry.name[0] == "!":
            raise RuntimeError(
                "plugin name %r in entry point group %r may not start with '!'"
                % (entry.name, entry_point_group)
            )
        try:
            pipeline_factory = entry.load().s
        except AttributeError:
            pipeline_factory = entry.load()
        loader.add_constructor(
            tag="!" + entry.name, constructor=yaml_constructor(pipeline_factory)
        )


def load_section_plugins(entry_point_group: str) -> Tuple[SectionPlugin]:
    """
    Load configuration plugins from an entry point group

    :param entry_point_group: entry point group to search
    :return: all loaded plugins
    """
    return tuple(
        SectionPlugin.load(entry_point)
        for entry_point in get_entrypoints(entry_point_group)
    )


@contextmanager
def load(config_path: str):
    """
    Load a configuration and keep it alive for the given context

    :param config_path: path to a configuration file
    """
    # we bind the config to _ to keep it alive
    if os.path.splitext(config_path)[1] in (".yaml", ".yml"):
        add_constructor_plugins(
            "cobald.config.yaml_constructors", COBalDLoader  # type: ignore
        )
        config_plugins = load_section_plugins("cobald.config.sections")
        _ = load_yaml_configuration(
            config_path,
            loader=COBalDLoader,  # type: ignore
            plugins=config_plugins,
        )
    elif os.path.splitext(config_path)[1] == ".py":
        _ = load_python_configuration(config_path)
    else:
        raise ValueError(
            "Unknown configuration extension: %r" % os.path.splitext(config_path)[1]
        )
    yield


def load_pipeline(content: list):
    """
    Load a cobald pipeline of Controller >> ... >> Pool from a configuration section

    :param content: content of the configuration section
    :return:
    """
    translator = PipelineTranslator()
    return translator.translate_hierarchy({"pipeline": content})


class PipelineTranslator(Translator):
    """
    Translator for :py:mod:`cobald` pipelines

    This allows for YAML configurations to have one or several ``pipeline`` elements.
    Each ``pipeline``  is translated as a series of nested elements, the way a
    :py:class:`~cobald.interfaces.Controller` receives a
    :py:class:`~cobald.interfaces.Pool`.

    .. code:: yaml

        pipeline:
            # same as ``package.module.callable(a, b, keyword1="one", keyword2="two")
            - __type__: package.module.Controller
              interval: 20
            - __type__: package.module.Pool
    """

    def translate_hierarchy(self, structure, *, where="", **construct_kwargs):
        try:
            pipeline = structure["pipeline"]
        except (KeyError, TypeError):
            return super().translate_hierarchy(
                structure, where=where, **construct_kwargs
            )
        else:
            prev_item, items = None, []
            for index, item in reversed(list(enumerate(pipeline))):
                if prev_item is not None:
                    try:
                        # fully constructed object from !constructor
                        prev_item = item >> prev_item
                    except TypeError:
                        # encoded object from __type__: constructor
                        prev_item = self.translate_hierarchy(
                            item, where="%s[%s]" % (where, index), target=prev_item
                        )
                else:
                    prev_item = self.translate_hierarchy(
                        item, where="%s[%s]" % (where, index)
                    )
                items.append(prev_item)
            return list(reversed(items))



from typing import Union

import trio

from ..interfaces import Pool, Controller
from ..utility import enforce, InvariantError, pairwise
from ..daemon import service


@service(flavour=trio)
class DemandSwitch(Controller):
    """
    Controller that dispatches to slaved controllers based on demand

    .. code:: python

        DemandSwitch(pool, linear_control, 10, supply_control)

    :param target: the pool on which to regulate demand
    :param default: controller to use by default
    :param slaves: pairs of minimum demand to switch and corresponding controller
    :param interval: interval between adjustments in seconds
    """

    def __init__(
        self,
        target: Pool,
        default: Controller,
        *slaves: Union[float, Controller],
        interval=1,
    ):
        super().__init__(target)
        enforce(
            len(slaves) % 2 == 0,
            InvariantError("slaves must be paired with required demands"),
        )
        self._default = default
        self._slaves = tuple(sorted(pairwise(slaves)))
        enforce(
            all(
                (isinstance(demand, (int, float)) and isinstance(slave, Controller))
                for demand, slave in self._slaves
            ),
            InvariantError("slaves must be paired with required demands (float, int)"),
        )
        enforce(
            default.target in (None, target)
            and all(slave.target in (None, target) for _, slave in self._slaves),
            InvariantError("slaves must have None or same target as switch target"),
        )
        default.target = target
        for _, slave in self._slaves:
            slave.target = target
        self.interval = interval

    async def run(self):
        while True:
            self.regulate_demand(self.interval)
            await trio.sleep(self.interval)

    def regulate(self, interval):
        chosen = self._default
        for demand, slave in self._slaves:
            if demand <= self.target.demand:
                chosen = slave
        chosen.regulate(interval)



from typing import Type, Tuple

from yaml import SafeLoader, BaseLoader, nodes

from .mapping import (
    load_configuration as load_mapping_configuration,
    ConfigurationError,
    SectionPlugin,
)


def load_configuration(
    path: str, loader: Type[BaseLoader] = SafeLoader, plugins: Tuple[SectionPlugin] = ()
):
    with open(path) as yaml_stream:
        loader_instance = loader(yaml_stream)
        try:
            config_data = loader_instance.get_single_data()
        finally:
            loader_instance.dispose()
    return load_mapping_configuration(config_data=config_data, plugins=plugins)


def yaml_constructor(factory):
    """
    Convert a factory function/class to a YAML constructor

    :param factory: the factory function/class
    :return: factory constructor

    Applying this helper to a factory allows it to be used as a YAML constructor,
    without it knowing about YAML itself.
    It properly constructs nodes and converts
    mapping nodes to ``factory(**node)``,
    sequence nodes to ``factory(*node)``, and
    scalar nodes to ``factory()``.

    For example, registering the constructor ``yaml_constructor(factory)`` as
    ``!factory`` means the following YAML is converted to ``factory(a=0.3, b=0.7)``:

    .. code::

        - !factory
          a: 0.3
          b: 0.7
    """

    def factory_constructor(loader: BaseLoader, node: nodes.Node):
        if isinstance(node, nodes.MappingNode):
            kwargs = loader.construct_mapping(node)
            return factory(**kwargs)
        elif isinstance(node, nodes.ScalarNode):
            return factory()
        elif isinstance(node, nodes.SequenceNode):
            args = loader.construct_sequence(node)
            return factory(*args)
        else:
            raise ConfigurationError(
                "YAML constructor %r on unsupported node type %s"
                % (node.tag, type(node).__name__)
            )

    return factory_constructor



from typing import Any
from types import ModuleType

from functools import partial, singledispatch


@singledispatch
def pretty_ref(obj: Any) -> str:
    """Pretty object reference using ``module.path:qual.name`` format"""
    try:
        return obj.__module__ + ":" + obj.__qualname__
    except AttributeError:
        return pretty_ref(type(obj)) + "(...)"


@pretty_ref.register(partial)
def pretty_partial(obj: partial) -> str:
    if not obj.args and not obj.keywords:
        return pretty_ref(obj.func)
    return "partial(%s%s%s)" % (
        pretty_ref(obj.func),
        "" if not obj.args else ", ".join(repr(arg) for arg in obj.args),
        ""
        if not obj.keywords
        else ", ".join("%r = %r" % (k, v) for k, v in obj.keywords.items()),
    )


@pretty_ref.register(ModuleType)
def pretty_module(obj: ModuleType) -> str:
    return obj.__name__


class NameRepr(object):
    """
    Lazy pretty formatter for name of objects
    """

    def __init__(self, target):
        self.target = target

    def __str__(self):
        target = self.target
        return pretty_ref(target)

    __repr__ = __str__



from typing import Callable
import weakref

import trio

from cobald.interfaces import Pool, CompositePool
from cobald.daemon import service


@service(flavour=trio)
class FactoryPool(CompositePool):
    """
    Composition that adds and removes pools to satisfy demand

    :param factory: a callable that produces a new :py:class:`~.Pool`
    :param interval: how often to adjust the number of children

    Adjustment uses two extensions that children must respond to adequately:

    * When spawned via ``factory()``, children shall already be
      set to their expected ``demand``.
    * When disabled via ``demand=0``, children shall shut down
      and free any resources and tasks.

    Once spawned, children are free to adjust their demand if required.
    A child may disable itself permanently by setting its own ``demand = 0``.
    The :py:class:`FactoryPool` inspects the demand for all its children
    before spawning or disabling any children.

    Any child which satisfies ``supply > 0`` **or** ``demand > 0`` is considered
    active and contributes to the :py:class:`FactoryPool`
    ``supply``, ``demand``, ``allocation``, and ``utilisation``.
    The :py:class:`FactoryPool` makes no assumption about the validity or fitness
    of active children.
    It is the responsibility of children to report their status accordingly.
    For example, if a child shuts down and does not allocate its ``supply`` further,
    it should scale its reported ``allocation`` accordingly.
    """

    @property
    def children(self):
        return [*self._hatchery, *self._mortuary]

    @property
    def demand(self):
        return self._demand

    @demand.setter
    def demand(self, value):
        # we may spend an arbitrary time spawning Drones,
        # just acknowledge demand and defer any actions
        self._demand = value

    @property
    def supply(self):
        return sum(child.supply for child in self.children)

    @property
    def utilisation(self):
        active_children = [child for child in self.children if child.supply > 0]
        try:
            return sum(child.utilisation for child in active_children) / len(
                active_children
            )
        except ZeroDivisionError:
            return 1.0

    @property
    def allocation(self):
        active_children = [child for child in self.children if child.supply > 0]
        try:
            return sum(child.allocation for child in active_children) / len(
                active_children
            )
        except ZeroDivisionError:
            return 1.0

    def __init__(
        self, *children: Pool, factory: Callable[[], Pool], interval: float = 30
    ):
        self._demand = sum(child.demand for child in children)
        #: children fulfilling our demand
        self._hatchery = set(children)
        #: children shutting down
        self._mortuary = weakref.WeakSet()
        self.factory = factory
        self.interval = interval

    async def run(self):
        while True:
            await trio.sleep(self.interval)
            # freeze target demand in case another thread updates us
            supply, demand = self.supply, self.demand
            if supply > demand:
                self._shrink(target=demand)
            else:
                self._grow(target=demand)

    def _shrink(self, target: float):
        # we can only reap children that are not already shutting down
        # prefer reaping children that supply few used resources
        hit_list = sorted(
            self._hatchery, key=lambda child: child.supply * child.utilisation
        )
        excess_demand = sum(child.demand for child in hit_list) - target
        for child in hit_list:
            if excess_demand <= 0:
                break
            # reap child
            if child.demand <= excess_demand:
                excess_demand -= child.demand
                self._release_child(child)
        self._reap_children()

    def _grow(self, target: float):
        missing_demand = target - sum(child.demand for child in self.children)
        while missing_demand > 0:
            new_child = self.factory()
            self._hatchery.add(new_child)
            assert (
                new_child.demand > 0
            ), "factory must produce children with initial demand"
            missing_demand -= new_child.demand
        self._reap_children()

    def _reap_children(self):
        for child in list(self._hatchery):
            if child.demand <= 0:
                self._release_child(child)

    def _release_child(self, child: Pool):
        child.demand = 0
        self._hatchery.discard(child)
        self._mortuary.add(child)



import abc
from typing import TypeVar, Type

from ._pool import Pool
from ._partial import Partial


C = TypeVar("C", bound="Controller")


class Controller(metaclass=abc.ABCMeta):
    """
    Controller adjusting the demand in a :py:class:`~.Pool`

    :param target: the resource pool for which demand is adjusted
    """

    def __init__(self, target: Pool):
        self.target = target

    @classmethod
    def s(cls: Type[C], *args, **kwargs) -> Partial[C]:
        """
        Create an unbound prototype of this class, partially applying arguments

        .. code:: python

            controller = Controller.s(interval=20)

            pipeline = controller(rate=10) >> pool
        """
        return Partial(cls, *args, __leaf__=False, **kwargs)



import argparse

CLI = argparse.ArgumentParser(description="COBalD - the Opportunistic Balancing Daemon")
CLI.add_argument("CONFIGURATION", help="path of the configuration to use", type=str)
CLI_LOG = CLI.add_argument_group("Startup Logging")
CLI_LOG.add_argument(
    "--log-level",
    help="initial logging level",
    default="INFO",
    choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    type=str.upper,
)
CLI_LOG.add_argument(
    "--log-target",
    help="initial logging target; stdout, stderr or a file path",
    default="stderr",
)
CLI_LOG.add_argument(
    "--log-journal",
    help="use short formatting suitable for journals",
    action="store_true",
)



import abc
from typing import List


from ._pool import Pool


class CompositePool(Pool):
    """
    Concatenation of multiple providers for a number of indistinguishable resources
    """

    @property
    @abc.abstractmethod
    def supply(self):
        """The volume of resources that is provided by this pool"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def demand(self):
        """The volume of resources to be provided by this pool"""
        raise NotImplementedError

    @demand.setter
    @abc.abstractmethod
    def demand(self, value):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def utilisation(self) -> float:
        """Fraction of the provided resources which are actively used"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def allocation(self) -> float:
        """Fraction of the provided resources which are assigned for usage"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def children(self) -> List[Pool]:
        """The individual resource providers making up this pool"""
        raise NotImplementedError

    @children.setter
    @abc.abstractmethod
    def children(self, value: List[Pool]):
        raise NotImplementedError



import logging
import weakref
import trio
import gc
import functools
import threading

from types import ModuleType

from .meta_runner import MetaRunner
from .guard import exclusive
from ..debug import NameRepr


class ServiceUnit(object):
    """
    Definition for running a service

    :param service: the service to run
    :param flavour: runner flavour to use for running the service
    """

    __active_units__ = weakref.WeakSet()  # type: weakref.WeakSet[ServiceUnit]

    def __init__(self, service, flavour):
        assert hasattr(service, "run"), "service must implement a 'run' method"
        assert any(flavour == runner.flavour for runner in MetaRunner.runner_types), (
            "service flavour must be one of %s"
            % ",".join(repr(runner.flavour) for runner in MetaRunner.runner_types)
        )
        self.service = weakref.ref(service)
        self.flavour = flavour
        self._started = False
        ServiceUnit.__active_units__.add(self)

    @classmethod
    def units(cls):
        """Sequence of all currently defined units"""
        return list(cls.__active_units__)

    @property
    def running(self):
        return self._started

    def start(self, runner: MetaRunner):
        service = self.service()
        if service is None:
            return
        else:
            self._started = True
            runner.register_payload(service.run, flavour=self.flavour)

    def __repr__(self):
        return "%s(%r, flavour=%r)" % (
            self.__class__.__name__,
            self.service() or "<defunct>",
            self.flavour,
        )


def service(flavour):
    r"""
    Mark a class as implementing a Service

    Each Service class must have a ``run`` method, which does not take any arguments.
    This method is :py:meth:`~.ServiceRunner.adopt`\ ed after the daemon starts, unless

    * the Service has been garbage collected, or
    * the ServiceUnit has been :py:meth:`~.ServiceUnit.cancel`\ ed.

    For each service instance, its :py:class:`~.ServiceUnit` is available at
    ``service_instance.__service_unit__``.
    """

    def service_unit_decorator(raw_cls):
        __new__ = raw_cls.__new__

        def __new_service__(cls, *args, **kwargs):
            if __new__ is object.__new__:
                self = __new__(cls)
            else:
                self = __new__(cls, *args, **kwargs)
            service_unit = ServiceUnit(self, flavour)
            self.__service_unit__ = service_unit
            return self

        raw_cls.__new__ = __new_service__
        if raw_cls.run.__doc__ is None:
            raw_cls.run.__doc__ = "Service entry point"
        return raw_cls

    return service_unit_decorator


class ServiceRunner(object):
    """
    Runner for coroutines, subroutines and services
    """

    def __init__(self, accept_delay: float = 1):
        self._logger = logging.getLogger("cobald.runtime.daemon.services")
        self._meta_runner = MetaRunner()
        self._must_shutdown = False
        self._is_shutdown = threading.Event()
        self.running = threading.Event()
        self.accept_delay = accept_delay

    def execute(self, payload, *args, flavour: ModuleType, **kwargs):
        """
        Synchronously run ``payload`` and provide its output

        If ``*args*`` and/or ``**kwargs`` are provided, pass them to ``payload``
        upon execution.
        """
        if args or kwargs:
            payload = functools.partial(payload, *args, **kwargs)
        return self._meta_runner.run_payload(payload, flavour=flavour)

    def adopt(self, payload, *args, flavour: ModuleType, **kwargs):
        """
        Concurrently run ``payload`` in the background

        If ``*args*`` and/or ``**kwargs`` are provided, pass them to ``payload``
        upon execution.
        """
        if args or kwargs:
            payload = functools.partial(payload, *args, **kwargs)
        self._meta_runner.register_payload(payload, flavour=flavour)

    @exclusive()
    def accept(self):
        """
        Start accepting synchronous, asynchronous and service payloads

        Since services are globally defined, only one :py:class:`ServiceRunner`
        may :py:meth:`accept` payloads at any time.
        """
        if self._meta_runner:
            raise RuntimeError("payloads scheduled for %s before being started" % self)
        self._must_shutdown = False
        self._logger.info("%s starting", self.__class__.__name__)
        # force collecting objects so that defunct,
        # migrated and overwritten services are destroyed now
        gc.collect()
        self._adopt_services()
        self.adopt(self._accept_services, flavour=trio)
        self._meta_runner.run()

    def shutdown(self):
        """Shutdown the accept loop and stop running payloads"""
        self._must_shutdown = True
        self._is_shutdown.wait()
        self._meta_runner.stop()

    async def _accept_services(self):
        delay, max_delay, increase = 0.0, self.accept_delay, self.accept_delay / 10
        self._is_shutdown.clear()
        self.running.set()
        try:
            self._logger.info("%s started", self.__class__.__name__)
            while not self._must_shutdown:
                self._adopt_services()
                await trio.sleep(delay)
                delay = min(delay + increase, max_delay)
        except Exception:
            self._logger.exception("%s aborted", self.__class__.__name__)
            raise
        else:
            self._logger.info("%s stopped", self.__class__.__name__)
        finally:
            self.running.clear()
            self._is_shutdown.set()

    def _adopt_services(self):
        for unit in ServiceUnit.units():  # type: ServiceUnit
            if unit.running:
                continue
            self._logger.info("%s adopts %s", self.__class__.__name__, NameRepr(unit))
            unit.start(self._meta_runner)



from ..interfaces import Pool, CompositePool


class WeightedComposite(CompositePool):
    """
    Weighted composition of several pools, with each pool weighted by its
    supply, utilisation or allocation.
    """

    children = []

    @property
    def demand(self):
        return self._demand

    @demand.setter
    def demand(self, value):
        self._demand = value
        child_count = len(self.children)
        for pool in self.children:
            try:
                pool.demand = value * getattr(pool, self._weight) / self._total_weight
            except ZeroDivisionError:
                pool.demand = value / child_count

    @property
    def supply(self):
        return sum(child.supply for child in self.children)

    @property
    def utilisation(self):
        try:
            return (
                sum(
                    child.utilisation * getattr(child, self._weight)
                    for child in self.children
                )
                / self._total_weight
            )
        except ZeroDivisionError:
            return 1.0

    @property
    def allocation(self):
        try:
            return (
                sum(
                    child.allocation * getattr(child, self._weight)
                    for child in self.children
                )
                / self._total_weight
            )
        except ZeroDivisionError:
            return 1.0

    @property
    def _total_weight(self):
        return sum(getattr(child, self._weight) for child in self.children)

    def __init__(self, *children: Pool, weight="supply"):
        assert weight in (
            "supply",
            "utilisation",
            "allocation",
        ), "weight must be either supply, utilisation or allocation"
        self._weight = weight
        self._demand = sum(child.demand for child in children)
        self.children = list(children)



from collections.abc import Mapping
from logging import Formatter, LogRecord
from typing import Dict, Set, Union, Any, TypeVar

from .format_json import RECORD_ATTRIBUTES


T = TypeVar("T")


def escape_key(key: str) -> str:
    assert isinstance(key, str)
    return key.replace(r",", r"\,").replace(r"=", r"\=").replace(r" ", r"\ ")


def escape_field(field: T) -> T:
    if isinstance(field, str):
        return '"' + field.replace("\\", r"\\").replace('"', r"\"") + '"'
    return field


def line_protocol(
    name, tags: dict = None, fields: dict = None, timestamp: float = None
) -> str:
    """
    Format a report as per InfluxDB line protocol

    :param name: name of the report
    :param tags: tags identifying the specific report
    :param fields: measurements of the report
    :param timestamp: when the measurement was taken, in **seconds** since the epoch
    """
    _escape_key = escape_key
    _escape_field = escape_field
    output_str = name.replace(r",", r"\,").replace(r" ", r"\ ")
    if tags:
        output_str += ","
        output_str += ",".join(
            "%s=%s" % (_escape_key(key), _escape_key(value))
            for key, value in sorted(tags.items())
        )
    output_str += " "
    output_str += ",".join(
        ("%s=%s" % (_escape_key(key), _escape_field(value))).replace("'", '"')
        for key, value in sorted(fields.items())
    )
    if timestamp is not None:
        # line protocol requires nanosecond precision, python uses seconds
        output_str += " %d" % (timestamp * 1e9)
    return output_str + "\n"


class LineProtocolFormatter(Formatter):
    """
    Formatter that emits data as InfluxDB Line Protocol

    :param tags: record data to use as tags
    :param resolution: resolution of timestamps in seconds

    The ``tags`` act as a whitelist for record keys if they are an iterable.
    When a dictionary is supplied, its values act as default values if the
    key is not in a record.

    The ``resolution`` allows summarising data by downsampling the timestamps
    to the given resolution, e.g. for a ``resolution`` of ``10`` you can expect
    timestamps 10, 20, 30, ...
    If ``resolution`` is ``None`` the timestamp is omitted from the Line Protocol
    and Telegraf will take care on setting the current timestamp.
    """

    def __init__(
        self,
        tags: Union[Dict[str, Any], Set[str], None] = None,
        resolution: float = None,
    ):
        super().__init__()
        self._default_tags = tags if isinstance(tags, Mapping) else {}
        self._tags_whitelist = set(tags) if tags is not None else set()
        self._fields_blacklist = self._tags_whitelist | set(RECORD_ATTRIBUTES)
        self._resolution = resolution

    def format(self, record: LogRecord) -> str:
        args = record.args
        if args == ({},):  # logger.info('message', {}) -> record.args == ({},)
            args = {}
        assert isinstance(
            args, Mapping
        ), "monitor record argument must be a mapping, not %r" % type(args)
        assert all(
            elem is not None for elem in args.values()
        ), "line protocol values must not be None"
        record.asctime = self.formatTime(record, self.datefmt)
        record.message = record.getMessage() if args else record.msg
        tags = self._default_tags.copy()
        tags.update(
            {key: value for key, value in args.items() if key in self._tags_whitelist}
        )
        fields = {
            key: value
            for key, value in args.items()
            if key not in self._fields_blacklist
        }
        timestamp = (
            record.created // self._resolution * self._resolution
            if self._resolution is not None
            else None
        )
        return line_protocol(
            name=record.message, tags=tags, fields=fields, timestamp=timestamp
        )


if __name__ == "__main__":
    import logging

    logger = logging.getLogger()
    logger.handlers = [logging.StreamHandler()]
    logger.handlers[0].formatter = LineProtocolFormatter(
        {"latitude": 49, "longitude": 8}
    )
    logger.warning("forecast", {"temperature": 298, "humidity": 0.45})



import asyncio
import threading
import sys

from .base_runner import BaseRunner
from .thread_runner import CapturingThread


async def awaitable_runner(runner: BaseRunner):
    """Execute a runner without blocking the event loop"""
    runner_thread = CapturingThread(target=runner.run)
    runner_thread.start()
    delay = 0.0
    while not runner_thread.join(timeout=0):
        await asyncio.sleep(delay)
        delay = min(delay + 0.1, 1.0)


def asyncio_main_run(root_runner: BaseRunner):
    """
    Create an ``asyncio`` event loop running in the main thread and watching runners

    Using ``asyncio`` to handle subprocesses requires a specific loop type
    to run in the main thread.
    This function sets up and runs the correct loop in a portable way.
    In addition, it runs a single :py:class:`~.BaseRunner` until completion
    or failure.

    .. seealso:: The `issue #8 <https://github.com/MatterMiners/cobald/issues/8>`_
                 for details.
    """
    assert (
        threading.current_thread() == threading.main_thread()
    ), "only main thread can accept asyncio subprocesses"
    if sys.platform == "win32":
        event_loop = asyncio.ProactorEventLoop()
        asyncio.set_event_loop(event_loop)
    else:
        event_loop = asyncio.get_event_loop()
        asyncio.get_child_watcher().attach_loop(event_loop)
    event_loop.run_until_complete(awaitable_runner(root_runner))



"""
Daemon core specific to cobald
"""
import sys
import logging
import platform

import cobald.__about__

from .logger import initialise_logging
from .cli import CLI
from .config import load
from .. import runtime


def run(configuration: str, level: str, target: str, short_format: bool):
    """Run the daemon and all its services"""
    initialise_logging(level=level, target=target, short_format=short_format)
    logger = logging.getLogger(__package__)
    logger.info("COBalD %s", cobald.__about__.__version__)
    logger.info(cobald.__about__.__url__)
    logger.info(
        "%s %s (%s)",
        platform.python_implementation(),
        platform.python_version(),
        sys.executable,
    )
    logger.debug(cobald.__about__.__file__)
    logger.info("Using configuration %s", configuration)
    with load(configuration):
        logger.info("Starting daemon services...")
        runtime.accept()


def cli_run():
    """Run the daemon from a command line interface"""
    options = CLI.parse_args()
    run(
        configuration=options.CONFIGURATION,
        level=options.log_level,
        target=options.log_target,
        short_format=options.log_journal,
    )



from functools import partial
from itertools import chain
from typing import Callable, Tuple, Optional, TypeVar, List, Set, Dict, overload

import trio

from ..interfaces import Pool, Controller, Partial
from ..daemon import service

C = TypeVar("C", bound="Controller")

#: Individual control rule for a pool on a given interval
#:
#: When a rule for a :py:class:`Stepwise` is invoked, it receives
#: the ``pool`` to manage and the ``interval`` elapsed since the
#: last modification.
#: It should either *return* the new :py:attr:`~.Pool.demand`, or
#: :py:const:`None` to indicate no change; the latter can also
#: mean that the function does not hit a ``return`` statement.
#:
#: .. py:function:: \ rule(pool: Pool, interval: float) -> Optional[float]
#:
#: Note that a rule should *not* modify the ``pool`` directly.
ControlRule = Callable[[Pool, float], Optional[float]]


class RangeSelector(object):
    """
    Container that stores rules for the range of their supply bounds

    :param base: base rule that has no lower bound
    :param rules: lower bound and its control rule
    """

    def __init__(self, base: ControlRule, *rules: Tuple[float, ControlRule]):
        self._lookup = self._compile_lookup(base, rules)

    def get_rule(self, supply: float):
        for (low, high), rule in self._lookup.items():
            if low <= supply < high:
                return rule

    @staticmethod
    def _compile_lookup(base, rules) -> Dict[Tuple[float, float], ControlRule]:
        if not rules:
            return {(0, float("inf")): base}
        lookup = {}
        thresholds, _rules = zip(*sorted(rules))
        for low, high, rule in zip(
            chain([0], thresholds),
            chain(thresholds, [float("inf")]),
            chain([base], _rules),
        ):
            if low == high:
                raise ValueError("Duplicate entries for threshold %s" % low)
            lookup[low, high] = rule
        return lookup


@service(flavour=trio)
class Stepwise(Controller):
    """
    Controller that selects from several strategies based on supply

    :see: :py:class:`UnboundStepwise` allows creating :py:class:`Stepwise` instances
          via decorators.
    """

    def __init__(
        self,
        target: Pool,
        base: ControlRule,
        *rules: Tuple[float, ControlRule],
        interval: float = 1,
    ):
        super().__init__(target)
        self.interval = interval
        self._selector = RangeSelector(base, *rules)

    async def run(self):
        target, interval = self.target, self.interval
        while True:
            current_rule = self._selector.get_rule(target.supply)
            demand = current_rule(target, interval)
            if demand is not None:
                self.target.demand = demand
            await trio.sleep(interval)


class UnboundStepwise(object):
    """
    Decorator interface for constructing a :py:class:`Stepwise` controller

    Apply this as a decorator to a :py:data:`~.ControlRule` callable to create
    a basic controller skeleton.
    The initial callable forms the base rule.
    Additional rules can be added for specific :py:attr:`~.Pool.supply` thresholds
    using :py:meth:`~.UnboundStepwise.add`.

    The skeleton can be used like a regular :py:class:`~.Controller`:
    calling it with a :py:class:`~.Pool` and update ``interval`` creates
    a :py:class:`~.Controller` instance with the given rules for the
    :py:class:`~.Pool`.

    .. code:: python

        # initial controller skeleton from base case
        @stepwise
        def control(pool: Pool, interval):
            return 10

        # additional rules above specific supply thresholds
        @control.add(supply=10)
        def quantized(pool: Pool, interval):
            if pool.utilisation < 0.5:
                return pool.supply - 1
            elif pool.allocation > 0.5:
                return pool.supply + 1

        @control.add(supply=100)
        def continuous(pool: Pool, interval):
            if pool.utilisation < 0.5:
                return pool.supply * 1.1
            elif pool.allocation > 0.5:
                return pool.supply * 0.9

        # create controller from skeleton
        pipeline = control(pool, interval=10)
    """

    def __init__(self, base: ControlRule):
        self.base = base
        self.rules = []  # type: List[Tuple[float, ControlRule]]
        self._thresholds = set()  # type: Set[float]

    @overload  # noqa: F811
    def add(self, rule: ControlRule, *, supply: float) -> ControlRule:
        ...

    @overload  # noqa: F811
    def add(self, rule: None, *, supply: float) -> Callable[[ControlRule], ControlRule]:
        ...

    def add(self, rule: ControlRule = None, *, supply: float):  # noqa: F811
        """
        Register a new rule above a given ``supply`` threshold

        Registration supports a single-argument form for use as a decorator,
        as well as a two-argument form for direct application.
        Use the former for ``def`` or ``class`` definitions,
        and the later for ``lambda`` functions and existing callables.

        .. code:: python

            @control.add(supply=10)
            def linear(pool, interval):
                if pool.utilisation < 0.75:
                    return pool.supply - interval
                elif pool.allocation > 0.95:
                    return pool.supply + interval

            control.add(
                lambda pool, interval: pool.supply * (
                    1.2 if pool.allocation > 0.75 else 0.9
                ),
                supply=100
            )
        """
        if supply in self._thresholds:
            raise ValueError("rule for threshold %s re-defined" % supply)
        if rule is not None:
            self.rules.append((supply, rule))
            self._thresholds.add(supply)
            return rule
        else:
            return partial(self.add, supply=supply)

    def s(self, *args, **kwargs) -> Partial[Stepwise]:
        """
        Create an unbound prototype of this class, partially applying arguments

        .. code:: python

            @stepwise
            def control(pool: Pool, interval):
                return 10

            pipeline = control.s(interval=20) >> pool

        :note: The partial rules are sealed, and :py:meth:`~.UnboundStepwise.add`
               cannot be called on it.
        """
        return Partial(Stepwise, self.base, *self.rules, *args, __leaf__=True, **kwargs)

    def __call__(self, target: Pool, interval: float = None):
        if interval is None:
            return Stepwise(target, self.base, *self.rules)
        return Stepwise(target, self.base, *self.rules, interval=interval)


stepwise = UnboundStepwise



import abc
from typing import TypeVar, Type

from ._partial import Partial


C = TypeVar("C", bound="Controller")


class Pool(metaclass=abc.ABCMeta):
    """
    Individual provider for a number of indistinguishable resources
    """

    @property
    @abc.abstractmethod
    def supply(self) -> float:
        """The volume of resources that is provided by this pool"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def demand(self) -> float:
        """The volume of resources to be provided by this pool"""
        raise NotImplementedError

    @demand.setter
    @abc.abstractmethod
    def demand(self, value: float):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def utilisation(self) -> float:
        """Fraction of the provided resources which are actively used"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def allocation(self) -> float:
        """Fraction of the provided resources which are assigned for usage"""
        raise NotImplementedError

    @classmethod
    def s(cls: Type[C], *args, **kwargs) -> Partial[C]:
        """
        Create an unbound prototype of this class, partially applying arguments

        .. code:: python

            pool = RemotePool.s(port=1337)

            pipeline = controller >> pool(host='localhost')
        """
        return Partial(cls, *args, __leaf__=True, **kwargs)



from cobald.interfaces import Pool, PoolDecorator

from ..utility import enforce
from ..utility.primitives import infinity as inf


class Standardiser(PoolDecorator):
    """
    Limits for changes to the demand of a pool

    :param target: the pool on which changes are standardised
    :param minimum: minimum ``target.demand`` allowed
    :param maximum: maximum ``target.demand`` allowed
    :param granularity: granularity of ``target.demand``
    :param surplus: maximum by which ``target.supply`` may exceed ``target.demand``
    :param backlog: maximum by which ``target.demand`` may exceed ``target.supply``

    The default values apply no limits at all so that isolated limits may be used.
    When several limits are set, ``granularity`` has the weakest priority,
    both ``surplus`` and ``backlog`` may limit the result of ``granularity``,
    and ``minimum`` and ``maximum`` overrule all other limits.
    It is illegal to
    """

    @property
    def demand(self) -> float:
        if abs(self._demand - self.target.demand) >= self.granularity:
            self._demand = self.target.demand
        return self._demand

    @demand.setter
    def demand(self, value: float):
        minimum = max(self.minimum, self.target.supply - self.backlog)
        maximum = min(self.maximum, self.target.supply + self.surplus)
        self._demand = type(value)(min(maximum, max(minimum, value)))
        request = value // self.granularity * self.granularity
        self.target.demand = type(value)(min(maximum, max(minimum, request)))

    def __init__(
        self,
        target: Pool,
        minimum: float = -inf,
        maximum: float = inf,
        granularity: int = 1,
        backlog: float = inf,
        surplus: float = inf,
    ):
        super().__init__(target)
        enforce(minimum <= maximum, ValueError("minimum must be smaller than maximum"))
        enforce(surplus > 0, ValueError("allowed surplus must be positive"))
        enforce(backlog > 0, ValueError("allowed backlog must be positive"))
        enforce(granularity > 0, ValueError("granularity must be positive"))
        # demand may be incrementally changed - store it internally to give
        # the impression of a smooth transition
        self._demand = target.demand
        self.minimum = minimum
        self.maximum = maximum
        self.granularity = granularity
        self.surplus = surplus
        self.backlog = backlog



import trio

from cobald.interfaces import Pool, Controller

from cobald.daemon import service


@service(flavour=trio)
class RelativeSupplyController(Controller):
    """
    Controller that adjusts demand relative to supply

    :param target: the pool to manage
    :param low_utilisation: pool utilisation below which resources are decreased
    :param high_allocation: pool allocation above which resources are increased
    :param low_scale: scale of ``target.supply`` when decreasing resources
    :param high_scale: scale of ``target.supply`` when increasing resources
    :param interval: interval between adjustments in seconds
    """

    def __init__(
        self,
        target: Pool,
        low_utilisation=0.5,
        high_allocation=0.5,
        low_scale=0.9,
        high_scale=1.1,
        interval=1,
    ):
        super().__init__(target=target)
        self.interval = interval
        assert low_utilisation <= high_allocation
        self.low_utilisation = low_utilisation
        self.high_allocation = high_allocation
        assert low_scale <= high_scale
        assert low_scale < 1
        assert high_scale > 1
        self.low_scale = low_scale
        self.high_scale = high_scale

    async def run(self):
        while True:
            self.regulate(self.interval)
            await trio.sleep(self.interval)

    def regulate(self, interval):
        if self.target.utilisation < self.low_utilisation:
            self.target.demand = self.target.supply * self.low_scale
        elif self.target.allocation > self.high_allocation:
            self.target.demand = self.target.supply * self.high_scale
        else:
            self.target.demand = self.target.supply



from collections.abc import Mapping
from logging import Formatter, LogRecord
import json


#: Attributes of a LogRecord.
# See <the docs `https://docs.python.org/3/library/logging.html#logrecord-attributes`>_.
RECORD_ATTRIBUTES = (
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "message",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
)


class JsonFormatter(Formatter):
    """
    Formatter that emits data as JSON

    :param fmt: default data for all records
    :param datefmt: format for timestamps

    The ``datefmt`` parameter has almost the same meaning as
   :py:class:`~.Formatter`.
    Setting it to ``None`` uses the default time format.
    However, setting it to any other value that is boolean
    false excludes the timestamp from reports.
    """

    def __init__(self, fmt: dict = None, datefmt: str = None):
        super().__init__(fmt=None, datefmt=datefmt, style="%")
        self._defaults = fmt or {}
        if not isinstance(self._defaults, Mapping):
            raise TypeError("`fmt` must be a Mapping or None")
        self._add_time = self.datefmt or self.datefmt is None

    def format(self, record: LogRecord):
        args = record.args
        if args == ({},):  # logger.info('message', {}) -> record.args == ({},)
            args = {}
        assert isinstance(
            args, Mapping
        ), "monitor record argument must be a mapping, not %r" % type(args)
        data = self._defaults.copy()
        if self._add_time:
            data["time"] = self.formatTime(record, self.datefmt)
        data["message"] = record.getMessage() if args else record.msg
        data.update(args)
        return json.dumps(data)


if __name__ == "__main__":
    import logging

    logger = logging.getLogger()
    logger.handlers = [logging.StreamHandler()]
    logger.handlers[0].formatter = JsonFormatter({"latitude": 49, "longitude": 8})
    logger.warning("forecast", {"temperature": 298, "humidity": 0.45})



from inspect import Signature, BoundArguments
from typing import Type, Generic, TypeVar, TYPE_CHECKING, Union, overload

from . import _pool

if TYPE_CHECKING:
    from ._controller import Controller
    from ._proxy import PoolDecorator
    from ._pool import Pool

    Owner = Union[Controller, PoolDecorator]
    C_co = TypeVar("C_co", bound=Owner)
else:
    Owner = Union[object]
    C_co = TypeVar("C_co")


class Partial(Generic[C_co]):
    r"""
    Partial application and chaining of Pool :py:class:`~.Controller`\ s
    and :py:class:`~.Decorator`\ s

    This class acts similar to :py:class:`functools.partial`,
    but allows for repeated application (currying) and
    explicit binding via the ``>>`` operator.

    .. code:: python

        # incrementally prepare controller parameters
        control = Partial(Controller, rate=10, interval=10)
        control = control(low_utilisation=0.5, high_allocation=0.9)

        # apply target by chaining
        pipeline = control >> Decorator() >> Pool()

    :note: The keyword argument ``__leaf__`` is reserved for internal usage.

    :note: Binding :py:class:`~.Controller`\ s and :py:class:`~.Decorator`\ s
           creates a temporary :py:class:`~.PartialBind`. Only binding to a
           :py:class:`~.Pool` as the last element creates a concrete binding.
    """
    __slots__ = ("ctor", "args", "kwargs", "leaf")

    def __init__(self, ctor: Type[C_co], *args, __leaf__, **kwargs):
        self.ctor = ctor
        self.args = args
        self.kwargs = kwargs
        self.leaf = __leaf__
        self._check_signature()

    def _check_signature(self):
        args, kwargs = self.args, self.kwargs
        if "target" in kwargs or (args and isinstance(args[0], _pool.Pool)):
            raise TypeError(
                "%s[%s] cannot bind 'target' by calling. "
                "Use `this >> target` instead." % (self.__class__.__name__, self.ctor)
            )
        try:
            if not self.leaf:
                args = None, *args
            _ = Signature.from_callable(self.ctor).bind_partial(
                *args, **kwargs
            )  # type: BoundArguments
        except TypeError as err:
            message = err.args[0]
            raise TypeError(
                "%s[%s] %s" % (self.__class__.__name__, self.ctor, message)
            ) from err

    def __call__(self, *args, **kwargs) -> "Partial[C_co]":
        return Partial(
            self.ctor, *self.args, *args, __leaf__=self.leaf, **self.kwargs, **kwargs
        )

    def __construct__(self, *args, **kwargs):
        return self.ctor(*args, *self.args, **kwargs, **self.kwargs)

    @overload  # noqa: F811
    def __rshift__(self, other: "Union[Owner, Pool, PartialBind[Pool]]") -> "C_co":
        ...

    @overload  # noqa: F811
    def __rshift__(self, other: "Union[Partial, PartialBind]") -> "PartialBind[C_co]":
        ...

    def __rshift__(self, other):  # noqa: F811
        if isinstance(other, PartialBind):
            return PartialBind(self, other.parent, *other.targets)
        elif isinstance(other, Partial):
            if other.leaf:
                return self >> other.__construct__()
            return PartialBind(self, other)
        else:
            return self.__construct__(other)

    def __repr__(self):
        return (
            "{self.__class__.__name__}(ctor={self.ctor.__name__}".format(self=self)
            + ", args={self.args}, kwargs={self.kwargs}".format(self=self)
            + ", leaf={self.leaf})".format(self=self)
        )


class PartialBind(Generic[C_co]):
    r"""
    Helper for recursively binding :py:class:`~.Controller`\ s
    and :py:class:`~.Decorator`\ s

    This helper is used to invert the operator precedence of ``>>``,
    allowing the last pair to be bound first.
    """
    __slots__ = ("parent", "targets")

    def __init__(
        self,
        parent: Partial[C_co],
        *targets: "Union[Partial[Owner], PartialBind[Owner]]",
    ):
        self.parent = parent
        self.targets = targets

    @overload  # noqa: F811
    def __rshift__(self, other: Partial[Owner]) -> "PartialBind[C_co]":
        ...

    @overload  # noqa: F811
    def __rshift__(self, other: "Pool") -> "C_co":
        ...

    def __rshift__(self, other: "Union[Pool, Partial[Owner]]"):  # noqa: F811
        if isinstance(other, _pool.Pool):
            pool = self.targets[-1] >> other
            for owner in reversed(self.targets[:-1]):
                pool = owner >> pool
            return self.parent >> pool
        elif isinstance(other, Partial) and other.leaf:
            return self >> other.__construct__()
        else:
            return PartialBind(self.parent, *self.targets, other)



import logging
import logging.config
import sys
from typing import Any, Dict, TypeVar, Callable, Tuple, Generic

from entrypoints import EntryPoint

_logger = logging.getLogger(__package__)


T = TypeVar("T")
#: type of a mapping element, matching JSON/YAML
M = TypeVar("M", str, int, float, bool, dict, list)


class ConfigurationError(Exception):
    def __init__(self, what: Any, where: str = None):
        self.where = where
        self.what = what
        super().__init__("invalid configuration element '%s': %s" % (where, what))


def configure_logging(logging_mapping: dict):
    _logger.info("Configuring logging")
    # > takes a default parameter, disable_existing_loggers, which defaults to True
    # > for reasons of backward compatibility. This may or may not be what you want
    # Note: this is *not* what we want, since we create several loggers in advance
    logging_mapping["disable_existing_loggers"] = logging_mapping.get(
        "disable_existing_loggers", False
    )
    logging.config.dictConfig(logging_mapping)


class Translator(object):
    """
    Translator from a mapping to an initialised object hierarchy
    """

    def translate_hierarchy(
        self, structure: M, *, where: str = "", **construct_kwargs
    ) -> M:
        try:
            if isinstance(structure, dict):
                structure = {
                    key: self.translate_hierarchy(value, where="%s.%s" % (where, key))
                    for key, value in structure.items()
                }
                if "__type__" in structure:
                    return self.construct(structure, **construct_kwargs)
                return structure
            elif isinstance(structure, list):
                # translate bottom up - need those lists to materialize
                # reversed and enumerate iterables
                return list(
                    reversed(
                        [
                            self.translate_hierarchy(
                                item, where="%s[%s]" % (where, index)
                            )
                            for index, item in reversed(list(enumerate(structure)))
                        ]
                    )
                )
            else:
                return structure
        except ConfigurationError as err:
            if err.where is None:
                raise ConfigurationError(what=err.what, where=where)
            raise
        except Exception as err:
            raise ConfigurationError(where=where, what=err)

    def construct(self, mapping: dict, **kwargs):
        """
        Construct an object from a mapping

        :param mapping: constructor definition, with ``__type__`` and keyword arguments
        :param kwargs: additional keyword arguments to pass to the constructor
        """
        assert "__type__" not in kwargs and "__args__" not in kwargs
        mapping = {**mapping, **kwargs}
        factory_fqdn = mapping.pop("__type__")
        factory = self.load_name(factory_fqdn)
        args = mapping.pop("__args__", [])
        return factory(*args, **mapping)

    @staticmethod
    def load_name(absolute_name: str):
        """Load an object based on an absolute, dotted name"""
        path = absolute_name.split(".")
        try:
            __import__(absolute_name)
        except ImportError:
            try:
                obj = sys.modules[path[0]]
            except KeyError:
                raise ImportError("No module named %r" % path[0])
            else:
                for component in path[1:]:
                    try:
                        obj = getattr(obj, component)
                    except AttributeError as err:
                        raise ConfigurationError(
                            what="no such object %r: %s" % (absolute_name, err)
                        )
                return obj
        else:  # ImportError is not raised if ``absolute_name`` points to a valid module
            return sys.modules[absolute_name]


class SectionPlugin(Generic[M]):
    __slots__ = "section", "digest", "required"

    __entry_point_flags__ = {"required"}

    def __init__(
        self, section: str, digest: Callable[[M], Any], required: bool = False
    ):
        self.section = section
        self.digest = digest
        self.required = required

    @classmethod
    def load(cls, entry_point: EntryPoint) -> "SectionPlugin":
        digest = entry_point.load()
        flags = set(entry_point.extras or [])
        if not flags <= cls.__entry_point_flags__:
            raise ValueError(
                "unrecognized config section option %r for entry pint %r"
                % (flags - cls.__entry_point_flags__, entry_point)
            )
        return cls(
            section=entry_point.name,
            digest=digest,
            **{option: option in flags for option in cls.__entry_point_flags__},
        )


def load_configuration(
    config_data: Dict[str, Any], plugins: Tuple[SectionPlugin] = ()
) -> Dict[SectionPlugin, Any]:
    try:
        logging_mapping = config_data.pop("logging")
    except KeyError:
        pass
    else:
        configure_logging(logging_mapping)
    # see if there is any unexpected config content
    unmatched = config_data.keys() - {plugin.section for plugin in plugins}
    if unmatched:
        raise ConfigurationError(
            where="root", what="unknown config sections %s" % ", ".join(unmatched)
        )
    content = {}
    for plugin in plugins:
        try:
            section_data = config_data[plugin.section]
        except KeyError:
            if plugin.required:
                raise ConfigurationError(
                    where="root", what="missing section %r" % plugin.section
                )
        else:
            # invoke the plugin and store possible output
            # to avoid it being garbage collected
            plugin_content = plugin.digest(section_data)
            if plugin_content is not None:
                content[plugin] = plugin_content
    return content



"""
===========================================
COBalD - the Opportunistic Balancing Daemon
===========================================

.. image:: https://readthedocs.org/projects/cobald/badge/?version=latest
    :target: http://cobald.readthedocs.io/en/latest/?badge=latest
    :alt: Documentation Status

This is a **draft** for a feedback based balancing system for providing
opportunistic resources.
"""
__title__ = "cobald"
__summary__ = "COBalD - the Opportunistic Balancing Daemon"
__url__ = "https://github.com/MatterMiners/cobald"

__version__ = "0.11.0"
__author__ = "Eileen Kuehn, Max Fischer"
__email__ = "mainekuehn@gmail.com"
__copyright__ = "2018 - 2019 %s" % __author__
__keywords__ = "opportunistic scheduling scheduler demand feedback-loop cobald"



class InvariantError(Exception):
    """An invariant is violated"""


def enforce(condition: bool, exception: BaseException = InvariantError()):
    """
    Enforce that ``condition`` is set by raising ``exception`` otherwise

    This is a replacement for ``assert`` statements as part of validation.
    It cannot be disabled with ``-O`` and may raise arbitrary exceptions.

    .. code:: Python

        def sqrt(value):
            condition(value > 0, ValueError('value must be greater than zero')
            return math.sqrt(value)
    """
    if not condition:
        raise exception


def pairwise(iterable):
    """Iterator yielding consecutive pairs from ``iterable``"""
    elements = iter(iterable)
    return zip(elements, elements)



import threading
import functools


def exclusive(via=threading.Lock):
    """
    Mark a callable as exclusive

    :param via: factory for a Lock to guard the callable

    Guards the callable against being entered again before completion.
    Explicitly raises a :py:exc:`RuntimeError` on violation.

    :note: If applied to a method, it is exclusive across all instances.
    """

    def make_exclusive(fnc):
        fnc_guard = via()

        @functools.wraps(fnc)
        def exclusive_call(*args, **kwargs):
            if fnc_guard.acquire(blocking=False):
                try:
                    return fnc(*args, **kwargs)
                finally:
                    fnc_guard.release()
            else:
                raise RuntimeError("exclusive call to %s violated")

        return exclusive_call

    return make_exclusive



from ._pool import Pool
from typing import TypeVar, Type


from ._partial import Partial


C = TypeVar("C", bound="PoolDecorator")


class PoolDecorator(Pool):
    """
    Decorator modifying how a pool provides resources

    :param target: the resource pool for which demand is adjusted
    """

    def __init__(self, target: Pool):
        self.target = target

    @classmethod
    def s(cls: Type[C], *args, **kwargs) -> Partial[C]:
        """
        Create an unbound prototype of this class, partially applying arguments

        .. code:: python

            decorator = Buffer.s(window=20)

            pipeline = controller >> decorator >> pool
        """
        return Partial(cls, *args, __leaf__=False, **kwargs)

    @property
    def supply(self):
        """The volume of resources that is provided by this site"""
        return self.target.supply

    @property
    def demand(self):
        """The volume of resources to be provided by this site"""
        return self.target.demand

    @demand.setter
    def demand(self, value):
        self.target.demand = value

    @property
    def utilisation(self) -> float:
        """Fraction of the provided resources which is actively used"""
        return self.target.utilisation

    @property
    def allocation(self) -> float:
        """Fraction of the provided resources which is assigned for usage"""
        return self.target.allocation



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

repo_base_dir = os.path.abspath(os.path.dirname(__file__))
# pull in the packages metadata
package_about = {}
with open(os.path.join(repo_base_dir, "src", "cobald", "__about__.py")) as about_file:
    exec(about_file.read(), package_about)


with open(os.path.join(repo_base_dir, "README.rst"), "r") as README:
    long_description = README.read()

TESTS_REQUIRE = ["pytest>=4.3.0", "pytest-timeout"]

if __name__ == "__main__":
    setup(
        name=package_about["__title__"],
        version=package_about["__version__"],
        description=package_about["__summary__"],
        long_description=long_description.strip(),
        author=package_about["__author__"],
        author_email=package_about["__email__"],
        url=package_about["__url__"],
        # >>> Source Code to distribute
        # root location for packages
        package_dir={"": "src"},
        # the 'cobald' top-level is not a namespace package,
        # we must explicitly point to it
        packages=[
            "cobald.%s" % pkg
            for pkg in find_packages(os.path.join(repo_base_dir, "src", "cobald"))
        ]
        + ["cobald"],
        py_modules=["cobald.__about__"],
        entry_points={
            "console_scripts": ["cobald = cobald.daemon.core.main:cli_run"],
            "cobald.config.yaml_constructors": [
                "%s = %s:%s" % (name, module, name)
                for name, module in (
                    ("LinearController", "cobald.controller.linear"),
                    ("RelativeSupplyController", "cobald.controller.relative_supply"),
                    ("Buffer", "cobald.decorator.buffer"),
                    ("Limiter", "cobald.decorator.limiter"),
                    ("Logger", "cobald.decorator.logger"),
                    ("Standardiser", "cobald.decorator.standardiser"),
                )
            ],
            "cobald.config.sections": [
                "pipeline = cobald.daemon.core.config:load_pipeline [required]"
            ],
        },
        # >>> Dependencies
        install_requires=["pyyaml", "trio>=0.4.0", "include", "entrypoints"],
        extras_require={
            "docs": ["sphinx", "sphinxcontrib-tikz", "sphinx_rtd_theme"],
            "test": TESTS_REQUIRE,
            "contrib": [
                "flake8",
                "flake8-bugbear",
                "black; implementation_name=='cpython'",
            ]
            + TESTS_REQUIRE,
        },
        # metadata for package search
        license="MIT",
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            "Development Status :: 2 - Pre-Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "Intended Audience :: Science/Research",
            "Intended Audience :: System Administrators",
            "Topic :: Adaptive Technologies",
            "Topic :: Office/Business :: Scheduling",
            "Topic :: System :: Distributed Computing",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
        keywords=package_about["__keywords__"],
        # unit tests
        setup_requires=["pytest-runner"],
        test_suite="cobald_tests",
        tests_require=TESTS_REQUIRE,
    )
